;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;;
;;;; PROCEDURES FOR AUTOMATICALLY GENERATING ENVIRONMENT DIAGRAMS
;;;;
;;;; This program was designed to help students (and TAs and instructors) to 
;;;; be able to quickly, easily, and automatically generate basic environment
;;;; diagrams.  It currently deals well with environments, simple bindings, 
;;;; and procedures.  In the future, it would be nice to have it actually
;;;; draw box-and-pointer diagrams for lists, vectors, and the like.
;;;;
;;;; For an example on how to use this package, see "env-dia-test.scm".
;;;;
;;;; by Gerald Dalley
;;;; Copyright (C) 2005
;;;; You may use this software as long as you don't package it up in something
;;;; that you're selling and you retain this copyright statement.  If you are 
;;;; an instructor and happen to use this for a class, please let me know at 
;;;; dalleyg@mit.edu (it's fun to hear about people using stuff I've 
;;;; written).  
;;;;
;;;; Go to the very bottom of this file for a list of methods that may be
;;;; called.  If you are looking through the code, a typical user should
;;;; focus on the procedures with names starting with "method/".  All other
;;;; procedures and bindings are for internal use only.
;;;;
;;;; Prerequisites:
;;;; 1) MIT Scheme.  The current version has been tested on the MIT 6.001 
;;;;    version of MIT Scheme 7.5.1.  Major modifications are necessary 
;;;;    to make this work on DrScheme.
;;;; 2) GraphViz.  Tested with version 2.2.1.  Does not work with version
;;;;    1.2 or earlier.  This is a free, open-source graph drawing/layout tool
;;;;    that does some of the grunt work for us.  Downloads are available
;;;;    at http://www.graphviz.org for Microsoft Windows, RedHat Linux, 
;;;;    Debian Linux, Mac OSX, FreeBSD, OpenBSD, Cygwin, and so on.  If you're
;;;;    really in a bind, the source is available as well.  Note that the 
;;;;    GraphViz executables must be in the current execution path.  If you 
;;;;    have multiple bindings in the same environment that point to the same
;;;;    object, you'll need GraphViz 2.3 (< 2.3 had a rendering bug).
;;;;
;;;; Change Log:
;;;;   2005-05-09 1424 Major structural changes--now encapsulates all of its
;;;;                   behavior in a single message-passing procedure.  This
;;;;                   avoids cluttering up the global environment.  It also
;;;;                   now shows compiled and primitive procedures and allows
;;;;                   for suppressing the rendering of undesired bindings.
;;;;   2005-04-21 0930 Tweaked prereqs to include version numbers
;;;;   2005-04-20 2327 Added header comments
;;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(define env-dia
  (let ()
    ;;;;;;; INTERNAL IMMUTABLE GLOBALS ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
    
    ;; Normal users should ignore these
    
    ;; Open TT style for pseudo-HTML
    (define *tt* "<font face=\"courier\">")
    ;; Close TT style for pseudo-HTML
    (define *tt-end* "</font>")
    ;; The base environment for env-dia (corresponds to the outermost 
    ;; let statement)
    (define *env-dia-environment* (the-environment))
    ;; Environment in which env-dia is defined
    (define *env-dia-parent* (environment-parent *env-dia-environment*))
    
    ;;;;;;; DEBUGGING SUPPORT ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
    
    ;; When false, debug output is suppressed
    (define *debug* #f)
    ;; Prints out a list of values if debug mode is on
    (define (debug . args) (if *debug* (for-each display args)))
    ;; Turns on debug mode
    (define (method/debug-on)  (set! *debug* #t))
    ;; Turns off debug mode
    (define (method/debug-off) (set! *debug* #f))
    ;; Indicates whether we're in debug mode
    (define (method/debug?) *debug*)
    
    ;;;;;;; RETAINED ENVIRONMENTS ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
    
    ;;; Retains a reference to environments that will retained for later display
    (define *retained-environments* '())
    
    ;;;; Call this procedure when you want to make sure some (temporary) 
    ;;;; environment gets retained so that it will show up in the environment
    ;;;; diagram.  Example:
    ;;;;    (define counter
    ;;;;      (let ((count 0))
    ;;;;        (lambda ()
    ;;;;          (set! count (+ 1 count))
    ;;;;          (env-dia 'retain (the-environment))
    ;;;;          count)))
    (define (method/retain-environment env)
      (debug "retaining " env "\n")
      (set! *retained-environments* (cons env *retained-environments*)))
    
    ;;;; Get rid of all environments saved with retain-environment
    (define (method/flush-retained-environments)
      (debug "flushing all retained environments\n")
      (set! *retained-environments* (list *env-dia-environment*)))
    
    ;;;;;;; BINDINGS ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
    
    ;; A list of (name frame) tuples that will be omitted from the diagram.
    ;; For example, in the global environment, the binding for ceiling
    ;; will never be shown (by default).
    (define *suppressed-names* 
      (append (list (list 'env-dia *env-dia-parent*))
              (map (lambda (name) (list name user-initial-environment))
                   '(* + - / < <= = > >= abs acos asin atan ceiling complex? 
                       cos dec denominator even? exact->inexact exact? exp 
                       expt floor gcd inc inexact->exact inexact? integer? 
                       lcm log max min modulo negative? number? numerator odd? 
                       positive? quotient rational? rationalize real? 
                       remainder round sin sqrt tan truncate zero? 
                       make-picture make-window pgm-file->picture 
                       picture->pgm-file picture-display picture-height 
                       picture-map picture-max picture-min picture-ref 
                       picture-width picture? procedure->picture 
                       show-window-size u6001-dir nil))))
    
    ;; Indicates that in the given frame, the binding for name (if it
    ;; exists) should be omitted from the diagram.
    (define (method/suppress-name name frame)
      (if (not (method/suppressed? name frame))
          (set! *suppressed-names* 
                (cons (list name frame) *suppressed-names*))))
    
    ;; Suppresses all (current) bindings in the given frame.  Useful if 
    ;; you wish to define a set of helper procedures that you don't want
    ;; to have show up in the results
    (define (method/suppress-frame frame)
      (for-each (lambda (name) (method/suppress-name name frame))
                (environment-bound-names frame)))
    
    ;; Make sure that the name in the given frame is no longer suppressed.
    ;; No action taken if that name was not suppressed in that frame.
    (define (method/unsuppress-name name frame)
      (define (finder entry) (and (eq? name (first entry))
                                  (equal? frame (second entry))))
      (if *debug* 
          (if (list-transform-positive *suppressed-names* finder)
              (debug "Unsuppressing " name " in frame " frame "\n")
              (debug "Could not unsuppress " name " in frame " frame "\n"))) 
      (set! *suppressed-names*
            (list-transform-negative *suppressed-names* finder)))
    
    ;; Unsuppresses all (current) names in a given frame
    (define (method/unsuppress-frame frame)
      (for-each (lambda (name) (method/unsuppress-name name frame))
                (environment-bound-names frame)))
    
    ;; Indicates whether a name in a given frame is currently suppressed from
    ;; output.
    (define (method/suppressed? name frame)
      (define (helper entries)
        (cond ((null? entries) #f)
              ((and (eq? name (caar entries)) (eq? frame (cadar entries)))
               #t)
              (else (helper (cdr entries)))))
      (helper *suppressed-names*))
    
    ;; Internal procedure for returning all non-suppressed bindings for 
    ;; some frame.
    (define (get-bindings frame)
      (debug "get-bindings: suppressed: " *suppressed-names* "\n")
      (list-transform-negative
       (environment-bindings frame)
       (lambda (binding) 
         (or 
          ;; Skip defunct bindings
          (null? (cdr binding))
          ;; Skip explicitly suppressed bindings
          (method/suppressed? (first binding) frame)))))    
    
    
    ;; Magic code from project 4 that determines whether a frame is the
    ;; global environment's frame.
    (define (global-environment? frame)
      (environment->package frame))
    
    ;; Retrieves the name of an environment.
    (define (env-name env)
      (if (global-environment? env) 'GLOBAL-ENVIRONMENT env))
    
    ;; Predicates for testing whether a binding's value is one of the various
    ;; supported procedure types
    (define (binding-is-compound? binding) 
      (compound-procedure? (cadr binding)))
    
    (define (binding-is-primitive? binding) 
      (primitive-procedure? (cadr binding)))
    
    (define (binding-is-compiled? binding) 
      (compiled-procedure? (cadr binding)))
    
    (define (binding-is-procedure? binding)
      (or (binding-is-compound? binding)
          (binding-is-primitive? binding)
          (binding-is-compiled? binding)))    
    
    ;;;;;;; STRINGIFICATION AND ESCAPING ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
    
    ;; List of required character escapings needed for GraphViz's pseudo-HTML
    (define *html-escape-sequences*
      '((#\< "&lt;") (#\> "&gt;") (#\" "&quot;") (#\& "&amp;")))
    
    ;; Take some object and generate an escaped string representation for it
    (define (html-escape obj)
      ; Build up string, doing any required character escaping
      (fold-right 
       (lambda (c acc)
         (let ((esc-entry (assq c *html-escape-sequences*)))
           (if esc-entry
               (string-append (second esc-entry) acc)
               (string-append (list->string `(,c)) acc))))
       ; Start with a null string
       ""
       ; Stringify
       (string->list (with-output-to-string (lambda () (write obj))))))
    
    ;; Hack to replace ampersand "&" characters with a semi-impossible 
    ;; left paren.  This should only be used for creating non-visible
    ;; port and object names.
    (define (dot-escape obj)
      (string-replace (html-escape obj) #\& #\())
    
    ;;;;;;; STRINGIFICATION AND ESCAPING ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
    
    ;; Max number of elements to show for the body of a procedure before using
    ;; ellipses.
    (define *max-body-parts* 3)
    
    ;;;;;;; ACTUAL OUTPUT METHODS AND PROCEDURES ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
    
    ;;;;
    ;;;; render-environment-diagram : string -> undefined
    ;;;;
    ;;;; Draws an environment diagram using all bindings reachable from
    ;;;; the global environment and all environments retained through calls
    ;;;; to retain-environment.  The base-filename should be a string indicating
    ;;;; the filename to use to output the file.  For example, to create
    ;;;; "foo.png", base-filename should be "foo".  Note that in the current
    ;;;; implementation, a temporary file "foo.dot" will also be created.  This
    ;;;; is the input file for GraphViz's dot program.  It does the graphical
    ;;;; layout for us.  The .dot file is not deleted.
    ;;;;
    ;;;; Future Work:
    ;;;; 1) Draw pairs, vectors, strings, etc. as objects.
    ;;;;    To do this, I might want to see if I can obtain the memory address of
    ;;;;    Scheme objects and use this as the node keys.
    ;;;; 2) Multiple output formats.
    ;;;;    This includes easy extensions such as using other output formats 
    ;;;;    supported by graphviz.  More ambitiously, it would be nice to see
    ;;;;    if I could output to either Visio's XML format and use Visio's
    ;;;;    layout engine (since graphviz doesn't do such a great job for some
    ;;;;    diagrams).
    ;;;; 3) Fix lambda bodies.
    ;;;;    I haven't spent enough time debugging the lambda bodies so that 
    ;;;;    ellipses and parentheses are all done correctly.  Currently there
    ;;;;    are a few cosmetic bugs.
    ;;;; 4) Port to DrScheme
    ;;;;
    (define (method/render-environment-diagram base-filename)
      (debug "render-environment-diagram for " base-filename "\n")
      
      ;; Shared file output port to avoid endlessly including it
      ;; in parameter lists
      (define f (open-output-file (string-append base-filename ".dot")))
      
      ;; Spits values out to the output port
      (define (->f . args) (for-each (lambda (x) (display x f)) args))
      
      ;; Tells whether the input object has already been output.  If #f is 
      ;; returned, assumes that the caller will output that object.  All future 
      ;; calls with the same object (in an eq? sense) will return #t.
      (define not-already-output?! 
        (let ((outputted-objects (list *env-dia-environment*)))
          (lambda (item)
            (debug "Checking if " item " has already been output.\n")
            (if (memq item outputted-objects)
                #f
                (begin
                  (set! outputted-objects (cons item outputted-objects))
                  #t)))))
      
      ;; Works like assq but on an alist with weakly-referenced keys.  
      ;; Automatically removes most alist elements with garbage-collected keys.
      ;;  TODO: There is probably a cleaner, yet still efficient alternative.
      (define (weak-assq key weak-list)
        (cond ((null? weak-list) 
               nil)
              ((eq? (weak-car (car weak-list)) key) 
               (car weak-list))
              ((weak-pair/car? (car weak-list)) 
               (weak-assq key (cdr weak-list)))
              ((and (not (null? (cdr weak-list)))
                    (weak-pair/car? (cadr weak-list)))
               (set-car! weak-list (cadr weak-list))
               (set-cdr! weak-list (cdr weak-list))
               (weak-assq key weak-list))
              (else 
               (weak-assq key (cdr weak-list)))))
      
      ;; Returns an identifier symbol for a given object.  Keeps a reference to
      ;; each object identified so that the same symbol is associated with that
      ;; object.  To avoid memory leaks, a weak reference system is used.
      (define get-identifier 
        (let ((next-ident 0)
              (objref->ident '()))
          (lambda (obj)
            ;; Look up the identifier
            (if (weak-assq obj objref->ident)
                ;; Make a symbol out of the found identifier
                (symbol-append 
                 'ident 
                 (string->symbol 
                  (number->string (weak-car (cdr (assq obj objref->ident))))))
                ;; Assign the identifier and then re-get it
                (begin
                  (set! objref->ident 
                        (cons (cons obj (weak-cons next-ident nil)) objref->ident))
                  (set! next-ident (inc next-ident))
                  (get-identifier obj))))))
      
      ;; Outputs a compound procedure (and its parent frame) 
      (define (render-compound-procedure proc)
        (if (not-already-output?! proc)
            (let* ((double-bubble (unsyntax proc))
                   (params        (if (eq? (first double-bubble) 'named-lambda)
                                      (cdr (second double-bubble))
                                      (second double-bubble)))
                   (body          (third double-bubble))
                   (parent-frame  (procedure-environment proc)))
              
              (debug "Creating double-bubble "
                     (if (eq? (first double-bubble) 'named-lambda)
                         (display (caadr double-bubble)))
                     "\n")
              
              ;; Draw double-bubble
              (->f "  \"" (get-identifier proc) "\" [label=<"
                   "<table border=\"1\" cellspacing=\"0\" cellpadding=\"0\" "
                   "bgcolor=\"#7777ff\">"
                   "<tr><td><table border=\"0\" cellspacing=\"0\" "
                   "cellpadding=\"0\">"
                   "<tr><td align=\"left\">P:</td><td align=\"left\">"
                   "" *tt* "" (html-escape params)
                   *tt-end* "</td></tr><tr><td align=\"left\">B:</td>"
                   "<td align=\"left\">" *tt* ""
                   (html-escape 
                    (if (<= (length body) *max-body-parts*)
                        body
                        (append (sublist body 0 *max-body-parts*) 
                                (list (string->symbol "...")))))
                   *tt-end* "</td></tr></table></td></tr></table>>];\n")
              
              ;; Link up to parent environment
              (->f "  \"" (get-identifier proc) "\" -> \"" 
                   (get-identifier (env-name parent-frame)) "\";\n\n")
              
              (render-frame parent-frame))))
      
      ;; Outputs a primitive procedure
      (define (render-primitive-procedure proc)
        (if (not-already-output?! proc)
            (begin
              (debug "Primitive " proc "\n")
              
              (->f " \"" (get-identifier proc) "\" [label=<"
                   "<table border=\"1\" cellspacing=\"0\" cellpadding=\"0\" "
                   "bgcolor=\"#77ff77\"><tr><td>"
                   (html-escape proc)
                   "</td></tr></table>"
                   ">];\n"))))
      
      ;; Outputs a compiled procedure
      (define (render-compiled-procedure proc)
        (if (not-already-output?! proc)
            (begin
              (debug "Compiled " proc "\n")
              
              (->f " \"" (get-identifier proc) "\" [label=<"
                   "<table border=\"1\" cellspacing=\"0\" cellpadding=\"0\" "
                   "bgcolor=\"#ff7777\"><tr><td>"
                   (html-escape proc)
                   "</td></tr></table>"
                   ">];\n"))))
      
      ;; Outputs a frame, its contents, and its parent frame
      ;;   TODO: handle pairs as separate objects
      ;;   TODO: handle vectors as separate objects
      (define (render-frame frame)
        (if (not-already-output?! frame)
            (let* ((bindings    (get-bindings frame))
                   (parent      (environment-parent frame))
                   (parent-name (env-name parent)))
              
              ;;; FRAME BLOCK
              
              (debug "Creating frame '" (env-name frame) "' with parent '"
                     parent-name "'\n")
              
              ;; Output frame header
              (->f "  \"" (get-identifier (env-name frame)) "\" [label=<"
                   "<table border=\"1\" cellborder=\"1\" cellspacing=\"0\" "
                   "cellpadding=\"0\">"
                   "<tr><td bgcolor=\"#CCccCC\">"
                   "<table border=\"0\"><tr><td align=\"left\">"
                   "" *tt* "" (html-escape (env-name frame)) 
                   *tt-end* "</td>"
                   "<td align=\"left\" port=\"name\"></td>"
                   "</tr></table>"
                   "</td></tr>")
              
              ;; Output bindings section if it exists 
              (if (not (null? bindings))
                  (begin
                    ;; Header
                    (->f "<tr><td><table border=\"1\" cellspacing=\"0\" "
                         "cellpadding=\"0\">")
                    
                    ;; Output all non-procedure bindings
                    (for-each 
                     (lambda (binding)
                       (->f "<tr><td><table border=\"0\"><tr><td align=\"left\">"
                            "" *tt* "" (html-escape (car binding))
                            *tt-end* "</td><td align=\"left\" port=\"" 
                            (html-escape (car binding)) "\">" *tt* 
                            (html-escape (cadr binding)) *tt-end* "</td>"
                            "</tr></table></td></tr>"))
                     (list-transform-negative bindings binding-is-procedure?))
                    
                    ;; Output all procedure bindings
                    (for-each 
                     (lambda (binding)
                       (->f "<tr><td><table border=\"0\"><tr><td align=\"left\">"
                            "" *tt* "" (html-escape (car binding))
                            *tt-end* "</td><td align=\"left\" port=\"" 
                            (get-identifier (car binding)) "\"></td>"
                            "</tr></table></td></tr>"))
                     (list-transform-positive bindings binding-is-procedure?))
                    
                    ;; Bindings footer
                    (->f "</table></td></tr>")))
              
              ;; Frame footer
              (->f "</table>>];\n")
              
              ;;; LINKS TO PROCEDURES
              
              (for-each
               (lambda (binding)
                 (->f "  \"" (get-identifier (env-name frame)) "\":\"" 
                      (get-identifier (car binding)) "\" -> "
                      "\"" (get-identifier (cadr binding)) "\";\n"))
               (list-transform-positive bindings binding-is-procedure?))
              
              ;;; LINK TO PARENT
              
              (if (not (global-environment? frame))
                  (->f "  \"" (get-identifier (env-name frame)) "\" -> \""
                       (get-identifier parent-name) "\";\n"))
              
              ;;; RECURSION
              
              (->f "\n")
              
              ;; render parent
              (if (not (global-environment? frame))
                  (render-frame (environment-parent frame)))
              
              ;; render procedures contained in this frame
              (for-each (lambda (binding) (render-compound-procedure (cadr binding)))
                        (list-transform-positive bindings binding-is-compound?))
              (for-each (lambda (binding) (render-primitive-procedure (cadr binding)))
                        (list-transform-positive bindings binding-is-primitive?))
              (for-each (lambda (binding) (render-compiled-procedure (cadr binding)))
                        (list-transform-positive bindings binding-is-compiled?))
              
              )))
      
      (define (render-header)
        (->f "digraph structs {\n"
             "  graph [rankdir=\"BT\"];\n"
             "    node [shape=plaintext];\n"
             "\n"))
      
      (define (render-footer)
        (->f "\n}\n"))    
      
      ; Start with the global environment (the parent of this
      ; procedure's frame).
      (render-header)
      (render-frame *env-dia-parent*)
      (for-each render-frame *retained-environments*)
      (render-footer)
      (close-port f)
      
      ;; Assumes GraphViz's dot program is in the current path and generates a 
      ;; PNG output file.
      (run-shell-command 
       (string-append "dot " base-filename ".dot -Tpng -o" 
                      base-filename ".png"))
      
      ;; At least on Windows, this will display the PNG (probably needs to be
      ;; commented out on *NIX systems).
      (run-shell-command
       (string-append base-filename ".png"))
      )
    
    ;;;;;;; DISPATCHER ;;;;;;;;;;;;;;;;;
    (lambda (msg . args)
      (case msg
        ;;; DEBUG MODE (displays lots of extra text)
        ;; Turn on debug mode (no args)
        ((debug-on)         (apply method/debug-on   args))
        ;; Turn off debug mode (no args)
        ((debug-off)        (apply method/debug-off  args))
        ;; Tell whether debug mode is enabled
        ((debug?)           (apply method/debug?     args))
        
        ;;; FRAME RETAINING (used to make sure that frames not reachable from 
        ;;;                  the global environment still get drawn)
        ;; Retain the supplied frame to make sure it shows up in the rendering
        ((retain)           (apply method/retain-environment args))
        ;; Forget all retained frames (no args)
        ((flush)            (apply method/flush-retained-environments args))
        
        ;;; BINDING SUPPRESSION (used to keep the environment diagram from 
        ;;;                      getting cluttered with bindings for built-in
        ;;;                      things such as nil, car, map, etc.  Can also
        ;;;                      manually add/remove items from the suppression
        ;;;                      list).
        ;; Takes a name (symbol) and frame and causes no binding to show up
        ;; for that name when rendering the supplied frame
        ((suppress-name)    (apply method/suppress-name args))
        ;; Suppresses all names currently bound in a given frame
        ((suppress-frame)   (apply method/suppress-frame args))
        ;; Removes suppression of a name (symbol) in a given frame.
        ((unsuppress-name)  (apply method/unsuppress-name args))
        ;; Removes all suppressions fora given frame.
        ((unsuppress-frame) (apply method/unsuppress-frame args))
        ;; Indicates whether a given name (symbol) is suppressed from rendering
        ;; in the supplied frame.
        ((suppressed?)      (apply method/suppressed args))
        
        ;;; RENDERING (used to draw the environment diagram as an image)
        ;; Takes a string that is the beginning of the filenames used for
        ;; the output rendering.  For example, if the argument is "foo", 
        ;; then the intermediate file "foo.dot" will be created and then
        ;; converted to "foo.png", which will be a rendering of the 
        ;; environment model for every (non-suppressed) binding reachable
        ;; from the global environment or any of the retained environments.
        ((render)           (apply method/render-environment-diagram args))
        
        ;;; EXTRA DEBUGGING SUPPORT 
        ;; The following methods, when uncommented, are useful for debugging.
        ;; The test method allows for calling any internal method directly.
        ;; The get method will return any binding in the env-dia internal
        ;; environment.  Normal users should not use these methods.
        ;((test)           (apply args))
        ;((get)            (environment-lookup (the-environment) (car args)))
        ))))
